# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.element import Element
from odfdo.header import Header
from odfdo.line_break import LineBreak
from odfdo.paragraph import Paragraph
from odfdo.tab import Tab


@pytest.fixture
def document(samples) -> Iterable[Document]:
    document = Document(samples("base_text.odt"))
    yield document


def test_get_paragraph_list(document):
    body = document.body
    paragraphs = body.get_paragraphs()
    assert len(paragraphs) == 7
    second = paragraphs[1]
    text = second.text
    assert text == "This is the second paragraph."


def test_empty_str(document):
    paragraph = Paragraph()
    assert (str(paragraph)) == "\n"


def test_get_paragraph_list_str(document):
    body = document.body
    paragraphs = body.get_paragraphs()
    second = paragraphs[1]
    assert (str(second)) == "This is the second paragraph.\n"


def test_get_paragraph_list_property(document):
    body = document.body
    paragraphs = body.paragraphs
    assert len(paragraphs) == 7
    second = paragraphs[1]
    text = second.text
    assert text == "This is the second paragraph."


def test_get_paragraph_list_style(document):
    body = document.body
    paragraphs = body.get_paragraphs(style="Hanging_20_indent")
    assert len(paragraphs) == 1
    paragraph = paragraphs[0]
    text = paragraph.text
    assert text == "This is a paragraph with a named style."


def test_get_paragraph_list_context(document):
    body = document.body
    section2 = body.get_section(position=1)
    paragraphs = section2.get_paragraphs()
    assert len(paragraphs) == 2
    paragraph = paragraphs[0]
    text = paragraph.text
    assert text == "First paragraph of the second section."


def test_get_paragraph_list_context_property(document):
    body = document.body
    section2 = body.get_section(position=1)
    paragraphs = section2.paragraphs
    assert len(paragraphs) == 2
    paragraph = paragraphs[0]
    text = paragraph.text
    assert text == "First paragraph of the second section."


def test_get_paragraph_by_content(document):
    body = document.body
    regex = r"(first|second|a) paragraph"
    paragraph = body.get_paragraph(content=regex)
    text = paragraph.text
    assert text == "This is the first paragraph."


def test_get_paragraph_by_content_context(document):
    body = document.body
    section2 = body.get_section(position=1)
    regex = r"([Ff]irst|second|a) paragraph"
    paragraph = section2.get_paragraph(content=regex)
    text = paragraph.text
    assert text == "First paragraph of the second section."


def test_odf_paragraph(document):
    body = document.body
    paragraph = body.get_paragraph()
    assert isinstance(paragraph, Paragraph)


def test_get_paragraph(document):
    body = document.body
    paragraph = body.get_paragraph(position=3)
    text = paragraph.text
    expected = "This is the first paragraph of the second title."
    assert text == expected


def test_insert_paragraph(document):
    body = document.body.clone
    paragraph = Paragraph("An inserted text", style="Text_20_body")
    body.append(paragraph)
    last_paragraph = body.get_paragraphs()[-1]
    assert last_paragraph.text == "An inserted text"


def test_insert_paragraph_property(document):
    body = document.body.clone
    paragraph = Paragraph("An inserted text", style="Text_20_body")
    body.append(paragraph)
    last_paragraph = body.paragraphs[-1]
    assert last_paragraph.text == "An inserted text"


def test_get_paragraph_missed(document):
    body = document.body
    paragraph = body.get_paragraph(position=999)
    assert paragraph is None


def test_create_naive_cr():
    para = Paragraph("\n")
    expected = "<text:p><text:line-break/></text:p>"
    assert para.serialize() == expected


def test_plain_text_splitted_length():
    txt = "A test,\n   \twith \n\n some é and \t and     5 spaces."
    para = Paragraph()
    content = para._expand_spaces(txt)
    content = para._merge_spaces(content)
    content = para._replace_tabs_lb(content)
    assert len(content) == 13


def test_plain_text_splitted_elements():
    txt = "A test,\n   \twith \n\n some é and \t and     5 spaces."
    para = Paragraph()
    content = para._expand_spaces(txt)
    content = para._merge_spaces(content)
    content = para._replace_tabs_lb(content)
    assert content[0] == "A test,"
    assert isinstance(content[1], LineBreak)
    assert isinstance(content[4], Tab)
    assert isinstance(content[6], LineBreak)


def test_plain_text_splitted_empty():
    para = Paragraph()
    content = para._expand_spaces("")
    content = para._merge_spaces(content)
    content = para._replace_tabs_lb(content)
    assert content == []


def test_append_plain_text_guess_utf8():
    txt = "A test,\n   \twith \n\n some \xe9 and \t and     5 spaces."
    para = Paragraph()
    para.append_plain_text(txt)
    expected = (
        '<text:p>A test,<text:line-break/> <text:s text:c="2"/>'
        "<text:tab/>with <text:line-break/><text:line-break/> "
        'some é and <text:tab/> and <text:s text:c="4"/>'
        "5 spaces.</text:p>"
    )
    assert para.serialize() == expected


def test_append_plain_text_guess_utf8_2():
    txt = "A test,\n   \twith \n\n some \xe9 and \t and     5 spaces."
    para = Paragraph()
    para.append(txt)
    expected = (
        '<text:p>A test,<text:line-break/> <text:s text:c="2"/>'
        "<text:tab/>with <text:line-break/><text:line-break/> "
        'some é and <text:tab/> and <text:s text:c="4"/>'
        "5 spaces.</text:p>"
    )
    assert para.serialize() == expected


def test_append_plain_text_guess_utf8_2_no_format():
    txt = "A test,\n   \twith \n\n some \xe9 and \t and     5 spaces."
    para = Paragraph()
    para.append(txt, formatted=False)
    expected = "<text:p>A test, with some é and and 5 spaces.</text:p>"
    assert para.serialize() == expected


def test_append_plain_text_guess_utf8_3_no_format():
    txt = " A test,\n   \twith \n\n some \xe9 and \t and     5 spaces. "
    para = Paragraph()
    para.append(txt, formatted=False)
    # old rule
    # expected = "<text:p> A test, with some é and and 5 spaces. </text:p>"
    # new rule, the result must comply with standard
    expected = (
        "<text:p><text:s/>A test, with some é and and 5 spaces.<text:s/></text:p>"
    )
    assert para.serialize() == expected


def test_append_plain_text_guess_utf8_4_no_format():
    txt = " A test,\n   \twith \n\n some \xe9 and \t and     5 spaces."
    para = Paragraph(" ")  # -> text:s text:c="1"/>
    para.append(txt, formatted=False)
    # old rule
    # expected = "<text:p> A test, with some é and and 5 spaces.</text:p>"
    expected = (
        '<text:p><text:s text:c="2"/>A test, with some é and and 5 spaces.</text:p>'
    )
    assert para.serialize() == expected


def test_append_plain_text_guess_utf8_5():
    txt = " A test,\n   \twith \n\n some \xe9 and \t and     5 spaces."
    para = Paragraph(" ")
    para.append(txt)
    expected = (
        '<text:p><text:s text:c="2"/>'
        'A test,<text:line-break/> <text:s text:c="2"/>'
        "<text:tab/>with <text:line-break/><text:line-break/> "
        'some é and <text:tab/> and <text:s text:c="4"/>'
        "5 spaces.</text:p>"
    )
    assert para.serialize() == expected


def test_append_simple():
    para = Paragraph("first")
    para.append(" second")
    assert para.text == "first second"


def test_append_simple_end_space():
    para = Paragraph("first ")
    para.append(" second")
    expected = "<text:p>first <text:s/>second</text:p>"
    assert para.serialize() == expected


def test_append_simple_end_2_space():
    para = Paragraph("first ")
    para.append("  second")
    expected = '<text:p>first <text:s text:c="2"/>second</text:p>'
    assert para.serialize() == expected


def test_append_simple_end_space_no_second():
    para = Paragraph("first ")
    para.append("second")
    assert para.text == "first second"


def test_append_multiline_text():
    para = Paragraph("first")
    para.append(" second\n third")
    assert para.text == "first second"
    assert para.inner_text == "first second\n third"


def test_append_multiline():
    para = Paragraph("first")
    para.append(" second\n third")
    expected = "<text:p>first second<text:line-break/> third</text:p>"
    assert para.serialize() == expected


def test_append_multiline_no_format():
    para = Paragraph("first")
    para.append(" second\n third", formatted=False)
    assert para.text == "first second third"


def test_append_plain_text_bytes_utf8():
    txt = "A test,\n   \twith \n\n some é and \t and     5 spaces."
    btxt: bytes = txt.encode("utf-8")
    para = Paragraph()
    para.append_plain_text(btxt)
    expected = (
        '<text:p>A test,<text:line-break/> <text:s text:c="2"/>'
        "<text:tab/>with <text:line-break/><text:line-break/> "
        'some é and <text:tab/> and <text:s text:c="4"/>'
        "5 spaces.</text:p>"
    )
    assert para.serialize() == expected


def test_append_plain_text_bytes_utf8_2():
    txt = "A test,\n   \twith \n\n some é and \t and     5 spaces."
    btxt: bytes = txt.encode("utf-8")
    para = Paragraph()
    para.append(btxt)
    expected = (
        '<text:p>A test,<text:line-break/> <text:s text:c="2"/>'
        "<text:tab/>with <text:line-break/><text:line-break/> "
        'some é and <text:tab/> and <text:s text:c="4"/>'
        "5 spaces.</text:p>"
    )
    assert para.serialize() == expected


def test_create_plain_text_utf8():
    txt = "A test,\n   \twith \n\n some é and \t and     5 spaces."
    para = Paragraph(txt)
    expected = (
        '<text:p>A test,<text:line-break/> <text:s text:c="2"/>'
        "<text:tab/>with <text:line-break/><text:line-break/> "
        'some é and <text:tab/> and <text:s text:c="4"/>'
        "5 spaces.</text:p>"
    )
    assert para.serialize() == expected


def test_create_no_plain_text_utf8():
    txt = "A test,\n   \twith \n\n some é and \t and     5 spaces."
    para = Paragraph(txt, formatted=False)
    result = para.serialize()
    assert "line-break" not in result
    assert "text:s" not in result
    assert "text:tab" not in result


def test_text():
    text = "Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_setter_1():
    text = "Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.text = None
    expected1 = "<text:p></text:p>"
    expected2 = "<text:p/>"
    assert paragraph.serialize() in (expected1, expected2)


def test_text_setter_2():
    text = "Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.text = "new text"
    expected = "<text:p>new text</text:p>"
    assert paragraph.serialize() == expected


def test_text_spaces_s1():
    text = " Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p><text:s/>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_s1_no_format():
    text = " Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text, formatted=False)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p><text:s/>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_s2():
    text = "  Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        '<text:p><text:s text:c="2"/>Le Père Noël a une moustache '
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_s2_no_format():
    text = "  Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text, formatted=False)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p><text:s/>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_s3():
    text = " \n Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p><text:s/><text:line-break/> Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_s3_no_format():
    text = " \n Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text, formatted=False)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p><text:s/>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_s4():
    text = " \n\t Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p><text:s/><text:line-break/><text:tab/> "
        "Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_s4_no_format():
    text = "<text:s/>\n\t Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text, formatted=False)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>&lt;text:s/&gt; Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_e1():
    text = "Le Père Noël a une moustache rouge. "
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "<text:s/></text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_e1_no_format():
    text = "Le Père Noël a une moustache rouge. "
    paragraph = Paragraph(text, formatted=False)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "<text:s/></text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_e2():
    text = "Le Père Noël a une moustache rouge.  "
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        '<text:s text:c="2"/>'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_e2_no_format():
    text = "Le Père Noël a une moustache rouge.  "
    paragraph = Paragraph(text, formatted=False)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "<text:s/></text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_e3():
    text = "Le Père Noël a une moustache rouge. \n "
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        " <text:line-break/><text:s/>"
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_e3_no_format():
    text = "Le Père Noël a une moustache rouge. \n "
    paragraph = Paragraph(text, formatted=False)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "<text:s/></text:p>"
    )
    assert paragraph.serialize() == expected


def test_multiple_append_1():
    paragraph = Paragraph("a")
    paragraph.append("b")
    paragraph.append("\n")
    paragraph.append("c")
    paragraph.append("d")
    paragraph.append(" ")
    expected = "<text:p>ab<text:line-break/>cd<text:s/></text:p>"
    assert paragraph.serialize() == expected


def test_multiple_append_2():
    paragraph = Paragraph("a")
    paragraph.append("b")
    paragraph.append("\n")
    paragraph.append("c")
    paragraph.append("d")
    paragraph.append(" ")
    paragraph.append(" ")
    expected = '<text:p>ab<text:line-break/>cd<text:s text:c="2"/></text:p>'
    assert paragraph.serialize() == expected


def test_init_append_element_paragraph():
    elem = Paragraph("a")
    paragraph = Paragraph(elem)
    expected = "<text:p>a<text:line-break/></text:p>"
    assert paragraph.serialize() == expected


def test_init_append_element_header():
    elem = Header(1, "header")
    paragraph = Paragraph(elem)
    expected = "<text:p>header<text:line-break/></text:p>"
    assert paragraph.serialize() == expected


def test_init_append_element_other():
    elem = Tab()
    paragraph = Paragraph(elem)
    expected = "<text:p><text:tab/></text:p>"
    assert paragraph.serialize() == expected


def test_text_spaces_se4():
    text = " \n Le Père Noël a une moustache rouge. \n\t "
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p><text:s/><text:line-break/> "
        "Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        " <text:line-break/><text:tab/>"
        "<text:s/></text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_spaces_se4_no_format():
    text = " \n Le Père Noël a une moustache rouge. \n\t "
    paragraph = Paragraph(text, formatted=False)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p><text:s/>Le Père Noël a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "<text:s/></text:p>"
    )
    assert paragraph.serialize() == expected


def test_tail():
    data = "<text:p>Le Père Noël a une <text:span>moustache</text:span> rouge.</text:p>"
    paragraph = Element.from_tag(data)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père Noël a une "
        "<text:span>moustache</text:span> "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_several():
    text = "Le Père rouge a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le Père "
        "<text:span "
        'text:style-name="highlight">rouge</text:span> '
        "a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_text_several_result():
    text = "Le Père rouge a une moustache rouge."
    paragraph = Paragraph(text)
    result = paragraph.set_span("highlight", regex="rouge")
    assert isinstance(result, list)
    assert len(result) == 2


def test_tail_several():
    data = (
        "<text:p>Le <text:span>Père</text:span> rouge a une moustache rouge.</text:p>"
    )
    paragraph = Element.from_tag(data)
    paragraph.set_span("highlight", regex="rouge")
    expected = (
        "<text:p>Le <text:span>Père</text:span> "
        "<text:span "
        'text:style-name="highlight">rouge</text:span> '
        "a une moustache "
        "<text:span "
        'text:style-name="highlight">rouge</text:span>.'
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_tail_several_result():
    data = (
        "<text:p>Le <text:span>Père</text:span> rouge a une moustache rouge.</text:p>"
    )
    paragraph = Element.from_tag(data)
    result = paragraph.set_span("highlight", regex="rouge")
    assert isinstance(result, list)
    assert len(result) == 2


def test_offset():
    text = "Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span("highlight", offset=text.index("moustache"))
    expected = (
        "<text:p>Le Père Noël a une "
        '<text:span text:style-name="highlight">moustache '
        "rouge.</text:span>"
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_offset_result():
    text = "Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    result = paragraph.set_span("highlight", offset=text.index("moustache"))
    assert isinstance(result, list)
    assert len(result) == 1


def test_offset_length():
    text = "Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span(
        "highlight", offset=text.index("moustache"), length=len("moustache")
    )
    expected = (
        "<text:p>Le Père Noël a une "
        '<text:span text:style-name="highlight">moustache'
        "</text:span> rouge."
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_set_span_none_regex():
    text = "Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    paragraph.set_span("highlight")
    expected = "<text:p>Le Père Noël a une moustache rouge.</text:p>"
    assert paragraph.serialize() == expected


def test_set_span_none_regex_result():
    text = "Le Père Noël a une moustache rouge."
    paragraph = Paragraph(text)
    result = paragraph.set_span("highlight")
    assert isinstance(result, list)
    assert len(result) == 0


def test_set_reference_mark_single(document):
    body = document.body
    para = body.get_paragraph()
    para.set_reference_mark("one", position=0)
    expected = (
        '<text:p text:style-name="Text_20_body">'
        '<text:reference-mark text:name="one"/>'
        "This is the first paragraph.</text:p>"
    )
    assert para.serialize() == expected


def test_set_reference_mark_single_2(document):
    body = document.body
    para = body.get_paragraph()
    para.set_reference_mark("one", position=2)
    expected = (
        '<text:p text:style-name="Text_20_body">'
        "Th"
        '<text:reference-mark text:name="one"/>'
        "is is the first paragraph.</text:p>"
    )
    assert para.serialize() == expected


def test_set_reference_mark_content(document):
    body = document.body
    para = body.get_paragraph()
    para.set_reference_mark("one", content=para)
    expected = (
        '<text:p text:style-name="Text_20_body">'
        '<text:reference-mark-start text:name="one"/>'
        "This is the first paragraph."
        '<text:reference-mark-end text:name="one"/>'
        "</text:p>"
    )
    assert para.serialize() == expected


def test_set_reference_mark_content_pos(document):
    body = document.body
    para = body.get_paragraph()
    para.set_reference_mark("one", position=(2, 4))
    expected = (
        '<text:p text:style-name="Text_20_body">'
        "Th"
        '<text:reference-mark-start text:name="one"/>'
        "is"
        '<text:reference-mark-end text:name="one"/>'
        " is the first paragraph."
        "</text:p>"
    )
    assert para.serialize() == expected


def test_set_reference_mark_content_2(document):
    body = document.body
    para = body.get_paragraph()
    para.set_reference_mark("one", content="first paragraph.")
    expected = (
        '<text:p text:style-name="Text_20_body">'
        "This is the "
        '<text:reference-mark-start text:name="one"/>'
        "first paragraph."
        '<text:reference-mark-end text:name="one"/>'
        "</text:p>"
    )
    assert para.serialize() == expected


def test_set_reference_mark_after(document):
    body = document.body
    para = body.get_paragraph()
    para.set_reference_mark("one", after="first")
    expected = (
        '<text:p text:style-name="Text_20_body">'
        "This is the first"
        '<text:reference-mark text:name="one"/>'
        " paragraph."
        "</text:p>"
    )
    assert para.serialize() == expected


def test_set_reference_mark_before(document):
    body = document.body
    para = body.get_paragraph()
    para.set_reference_mark("one", before="first")
    expected = (
        '<text:p text:style-name="Text_20_body">'
        "This is the "
        '<text:reference-mark text:name="one"/>'
        "first paragraph."
        "</text:p>"
    )
    assert para.serialize() == expected


def test_paragraph_formatted():
    p = Paragraph()
    p.text = "content"
    assert p.get_formatted_text() == "content\n\n"


def test_paragraph_formatted_simple():
    p = Paragraph()
    p.text = "content"
    assert p.get_formatted_text(simple=True) == "content"


def test_paragraph_formatted_context_simple():
    p = Paragraph()
    p.text = "content"
    context = {
        "document": None,
        "footnotes": [],
        "endnotes": [],
        "annotations": [],
        "rst_mode": False,
        "img_counter": 0,
        "images": [],
        "no_img_level": 0,
    }
    assert p.get_formatted_text(context, simple=True) == "content"



#!/usr/bin/env python
"""Each line of the text becomes a slide of the presentation, we change of style
depending on the length of text line.
"""

import os
from pathlib import Path

from odfdo import Document, DrawPage, Frame, Style

_DOC_SEQUENCE = 287
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "styled_prez"
TARGET = "presentation.odp"

CONTENT = """123
azertyuiop
azertyuiop azertyuiop
azertyuiop azertyuiop azertyuiop
azertyuiop azertyuiop azertyuiop azertyuiop
azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop
azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop
azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop
azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop
azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop azertyuiop
end.
""".splitlines()

# If text length is bigger then first value, use second value as font size:
TEXT_LEN_FONT_SIZE = [
    (95, 10),
    (80, 11),
    (65, 14),
    (50, 16),
    (40, 20),
    (30, 24),
    (20, 32),
    (10, 40),
    (5, 44),
    (-1, 48),
]


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def create_base_style() -> Style:
    """Creating a smooth style for the graphic item."""
    base_style = Style(
        "graphic",
        name="Gloup48",
        parent="standard",
        stroke="none",
        fill_color="#b3b3b3",
        textarea_vertical_align="middle",
        padding_top="1cm",
        padding_bottom="1cm",
        padding_left="1cm",
        padding_right="1cm",
        line_distance="0cm",
        guide_overhang="0cm",
        guide_distance="0cm",
    )
    base_style.set_properties(area="paragraph", align="center")
    base_style.set_properties(
        area="text",
        color="#dd0000",
        text_outline="false",
        font="Liberation Sans",
        font_family="Liberation Sans",  # compatibility
        font_style_name="Bold",
        family_generic="swiss",
        size="48pt",
        weight="bold",
    )
    return base_style


def add_styles(document: Document) -> None:
    """Generate all styles usable by the presentation as variations of a
    base style."""
    base_style = create_base_style()
    for _, font_size in TEXT_LEN_FONT_SIZE:
        variant_style: Style = base_style.clone
        variant_style.set_attribute("style:name", f"Gloup{font_size}")
        variant_style.set_properties(area="text", size=f"{font_size}pt")
        document.insert_style(variant_style)


def generate_document() -> Document:
    """Generate a Presentation Document with different styles."""
    presentation = Document("presentation")
    body = presentation.body
    body.clear()

    add_styles(presentation)

    for count, blurb in enumerate(CONTENT):
        text = blurb
        name = f"{count + 1} - {text[:10]}"
        page = DrawPage(name)
        # choosing some style:
        for text_len, font_size in TEXT_LEN_FONT_SIZE:
            if len(text) > text_len:
                size = font_size
                break
        style_name = f"Gloup{size}"

        text_frame = Frame.text_frame(
            text,
            size=("24cm", "2cm"),
            position=("2cm", "8cm"),
            style=style_name,
            text_style=style_name,
        )

        page.append(text_frame)
        body.append(page)

    return presentation


def main() -> None:
    document = generate_document()
    test_unit(document)
    save_new(document, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    body = document.body
    count = len([item for item in body.children if isinstance(item, DrawPage)])
    assert count == len(CONTENT)
    first_page = body.children[0]
    assert str(first_page).strip() == CONTENT[0].strip()
    last_page = body.children[-1]
    assert str(last_page).strip() == CONTENT[-1].strip()


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>

from collections.abc import Iterable
from datetime import datetime

import pytest

from odfdo.document import Document
from odfdo.user_field import UserDefined


@pytest.fixture
def document(samples) -> Iterable[Document]:
    document = Document(samples("meta.odt"))
    yield document


def test_user_defined_class():
    field = UserDefined()
    assert isinstance(field, UserDefined)


def test_user_defined_style():
    field = UserDefined(style="some_style")
    assert field.style == "some_style"


def test_create_user_defined_1(document):
    element = UserDefined(
        "unknown_in_meta",
        value=42,
        value_type="float",
        text=None,
        style=None,
        from_document=document,
    )
    expected = (
        '<text:user-defined text:name="unknown_in_meta" '
        'office:value-type="float" calcext:value-type="float" '
        'office:value="42" calcext:value="42">42'
        "</text:user-defined>"
    )
    assert element.serialize() == expected


def test_create_user_defined_2(document):
    element = UserDefined(
        "unknown_in_meta2",
        value=datetime(2013, 12, 30),
        value_type="date",
        text="2013-12-30",
        style=None,
        from_document=document,
    )
    expected = (
        '<text:user-defined text:name="unknown_in_meta2" '
        'office:value-type="date" calcext:value-type="date" '
        'office:date-value="2013-12-30T00:00:00">2013-12-30'
        "</text:user-defined>"
    )
    assert element.serialize() == expected


def test_create_user_defined_2_no_doc():
    element = UserDefined(
        "unknown_in_meta2",
        value=datetime(2013, 12, 30),
        value_type="date",
        text="2013-12-30",
        style=None,
        from_document=None,
    )
    expected = (
        '<text:user-defined text:name="unknown_in_meta2" '
        'office:value-type="date" calcext:value-type="date" '
        'office:date-value="2013-12-30T00:00:00">2013-12-30'
        "</text:user-defined>"
    )
    assert element.serialize() == expected


def test_create_user_defined_3_existing(document):
    element = UserDefined("Référence", from_document=document)
    expected = (  # noqa: UP031
        '<text:user-defined text:name="%s" '
        'office:value-type="boolean" calcext:value-type="boolean" '
        'office:boolean-value="true">true</text:user-defined>'
    ) % "Référence"
    assert element.serialize() == expected


def test_create_user_defined_4_existing(document):
    element = UserDefined(
        "Référence",
        value=False,  # default value if not existing
        value_type="boolean",
        from_document=document,
    )
    expected = (  # noqa: UP031
        '<text:user-defined text:name="%s" '
        'office:value-type="boolean" calcext:value-type="boolean" '
        'office:boolean-value="true">true</text:user-defined>'
    ) % "Référence"
    assert element.serialize() == expected


def test_create_user_defined_5_nodoc():
    element = UserDefined(
        "Référence",
        value=False,  # default value if not existing
        value_type="boolean",
        from_document=None,
    )
    expected = (  # noqa: UP031
        '<text:user-defined text:name="%s" '
        'office:value-type="boolean" calcext:value-type="boolean" '
        'office:boolean-value="false">false</text:user-defined>'
    ) % "Référence"
    assert element.serialize() == expected


def test_get_user_defined(document):
    element = UserDefined(
        "Référence",
        value=False,  # default value if not existing
        value_type="boolean",
        from_document=document,
    )
    body = document.body
    para = body.get_paragraph()
    para.append(element)
    user_defined = body.get_user_defined("Référence")
    expected = (  # noqa: UP031
        '<text:user-defined text:name="%s" '
        'office:value-type="boolean" calcext:value-type="boolean" '
        'office:boolean-value="true">true</text:user-defined>'
    ) % "Référence"
    assert user_defined.serialize() == expected


def test_get_user_defined_list(document):
    element = UserDefined(
        "Référence",
        value=False,  # default value if not existing
        value_type="boolean",
        from_document=document,
    )
    body = document.body
    para = body.get_paragraph()
    para.append(element)
    element2 = UserDefined(
        "unknown_in_meta2",
        value=datetime(2013, 12, 30),
        value_type="date",
        text="2013-12-30",
        style=None,
        from_document=None,
    )
    para.append(element2)
    user_defined_list = body.get_user_defined_list()
    assert len(user_defined_list) == 2


def test_get_user_defined_list_property(document):
    element = UserDefined(
        "Référence",
        value=False,  # default value if not existing
        value_type="boolean",
        from_document=document,
    )
    body = document.body
    para = body.get_paragraph()
    para.append(element)
    element2 = UserDefined(
        "unknown_in_meta2",
        value=datetime(2013, 12, 30),
        value_type="date",
        text="2013-12-30",
        style=None,
        from_document=None,
    )
    para.append(element2)
    user_defined_list = body.user_defined_list
    assert len(user_defined_list) == 2


def test_get_user_defined_value(document):
    element = UserDefined(
        "Référence",
        value=True,  # default value if not existing
        value_type="boolean",
        from_document=document,
    )
    body = document.body
    para = body.get_paragraph()
    para.append(element)
    element2 = UserDefined(
        "unknown_in_meta2",
        value=datetime(2013, 12, 30),
        value_type="date",
        text="2013-12-30",
        style=None,
        from_document=None,
    )
    para.append(element2)
    value = body.get_user_defined_value("Référence")
    assert value is True
    value = body.get_user_defined_value("unknown_in_meta2")
    assert value == datetime(2013, 12, 30)


def test_get_user_defined_get_none(document):
    body = document.body
    value = body.get_user_defined_value("unknown")
    assert value is None



#!/usr/bin/env python
"""Save a document as BytesIO."""

import io
import os
from pathlib import Path

from odfdo import Document, Paragraph

_DOC_SEQUENCE = 605
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "bytes"
TARGET = "document.odt"


def make_document() -> Document:
    """Return a Hello World document."""
    document = Document("text")
    body = document.body
    paragraph = Paragraph("Hello World")
    body.append(paragraph)
    return document


def document_to_bytesio(document: Document) -> bytes:
    with io.BytesIO() as bytes_content:
        document.save(bytes_content)
        # Now use the BytesIO in some way:
        # In a netwotk context, typically:
        #    response.write(bytes_content.getvalue())
        return bytes_content.getvalue()


def write_content(content: bytes) -> None:
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    path = OUTPUT_DIR / TARGET
    path.write_bytes(content)


def main() -> None:
    document = make_document()
    bytes_content = document_to_bytesio(document)
    test_unit(bytes_content)
    write_content(bytes_content)


def test_unit(content: bytes) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    with io.BytesIO() as bytes_content:
        bytes_content.write(content)
        bytes_content.seek(0)
        document = Document(bytes_content)
        assert document.body.search("Hello World") is not None


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal

import pytest

from odfdo.datatype import Boolean, Date, DateTime, Duration, Unit


def test_datetime_encode():
    date_value = datetime(2009, 0o6, 26, 11, 9, 36)
    expected = "2009-06-26T11:09:36"
    assert DateTime.encode(date_value) == expected


def test_datetime_encode_micro():
    date_value = datetime(2009, 6, 26, 11, 9, 36, 123456)
    expected = "2009-06-26T11:09:36.123456"
    assert DateTime.encode(date_value) == expected


def test_datetime_encode_UTC():
    date_value = datetime(2009, 6, 26, 11, 9, 36, tzinfo=timezone.utc)
    expected = "2009-06-26T11:09:36Z"
    assert DateTime.encode(date_value) == expected


def test_datetime_encode_micro_UTC():
    date_value = datetime(2009, 6, 26, 11, 9, 36, 123456, tzinfo=timezone.utc)
    expected = "2009-06-26T11:09:36.123456Z"
    assert DateTime.encode(date_value) == expected


def test_datetime_encode_gmt2():
    date_value = datetime(2009, 6, 26, 11, 9, 36, tzinfo=timezone(timedelta(hours=2)))
    expected = "2009-06-26T11:09:36+02:00"
    assert DateTime.encode(date_value) == expected


def test_datetime_encode_gmt_6():
    date_value = datetime(2009, 6, 26, 11, 9, 36, tzinfo=timezone(timedelta(hours=-6)))
    expected = "2009-06-26T11:09:36-06:00"
    assert DateTime.encode(date_value) == expected


def test_datetime_decode():
    date_value = "2009-06-29T14:33:21"
    expected = datetime(2009, 6, 29, 14, 33, 21)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_micro():
    date_value = "2009-06-29T14:33:21.123456"
    expected = datetime(2009, 6, 29, 14, 33, 21, 123456)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_nano():
    date_value = "2009-06-29T14:33:21.123456789"
    expected = datetime(2009, 6, 29, 14, 33, 21, 123456)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_utc():
    date_value = "2009-06-29T14:33:21Z"
    expected = datetime(2009, 6, 29, 14, 33, 21, tzinfo=timezone.utc)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_utc_00():
    date_value = "2009-06-29T14:33:21+00:00"
    expected = datetime(2009, 6, 29, 14, 33, 21, tzinfo=timezone.utc)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_gmt2():
    date_value = "2009-06-29T14:33:21+02:00"
    expected = datetime(2009, 6, 29, 14, 33, 21, tzinfo=timezone(timedelta(hours=2)))
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_gmt_6():
    date_value = "2009-06-29T14:33:21-06:00"
    expected = datetime(2009, 6, 29, 14, 33, 21, tzinfo=timezone(timedelta(hours=-6)))
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_utc_micro():
    date_value = "2009-06-29T14:33:21.123456Z"
    expected = datetime(2009, 6, 29, 14, 33, 21, 123456, tzinfo=timezone.utc)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_utc_00_micro():
    date_value = "2009-06-29T14:33:21.123456+00:00"
    expected = datetime(2009, 6, 29, 14, 33, 21, 123456, tzinfo=timezone.utc)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_gmt2_micro():
    date_value = "2009-06-29T14:33:21.123456+02:00"
    expected = datetime(
        2009, 6, 29, 14, 33, 21, 123456, tzinfo=timezone(timedelta(hours=2))
    )
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_utc_nano():
    date_value = "2009-06-29T14:33:21.123456789Z"
    expected = datetime(2009, 6, 29, 14, 33, 21, 123456, tzinfo=timezone.utc)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_utc_00_nano():
    date_value = "2009-06-29T14:33:21.123456789+00:00"
    expected = datetime(2009, 6, 29, 14, 33, 21, 123456, tzinfo=timezone.utc)
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_gmt2_nano():
    date_value = "2009-06-29T14:33:21.123456789+02:00"
    expected = datetime(
        2009, 6, 29, 14, 33, 21, 123456, tzinfo=timezone(timedelta(hours=2))
    )
    assert DateTime.decode(date_value) == expected


def test_datetime_decode_raises():
    date_value = "XXXXXX2009-06-29T14:33:21.123456789+02:00"
    with pytest.raises(ValueError):
        DateTime.decode(date_value)


def test_date_decode():
    date_value = "1999-12-25"
    expected = datetime(1999, 12, 25)
    assert Date.decode(date_value) == expected


def test_date_encode_1():
    date_value = date(1999, 12, 25)
    expected = "1999-12-25"
    assert Date.encode(date_value) == expected


def test_date_encode_2():
    date_value = datetime(1999, 12, 25, 0, 0, 0)
    expected = "1999-12-25"
    assert Date.encode(date_value) == expected


def test_duration_encode():
    duration = timedelta(0, 53, 0, 0, 6)
    expected = "PT00H06M53S"
    assert Duration.encode(duration) == expected


def test_duration_encode_raises():
    duration = []
    with pytest.raises(TypeError):
        Duration.encode(duration)


def test_duration_encode_neg():
    duration = timedelta(hours=-2)
    expected = "-PT02H00M00S"
    assert Duration.encode(duration) == expected


def test_duration_decode():
    duration = "PT12H34M56S"
    expected = timedelta(0, 56, 0, 0, 34, 12)
    assert Duration.decode(duration) == expected


def test_duration_decode_neg():
    duration = "-PT02H00M00S"
    expected = timedelta(hours=-2)
    assert Duration.decode(duration) == expected


def test_duration_decode_days():
    duration = "PT01D02H00M00S"
    expected = timedelta(days=1, hours=2)
    assert Duration.decode(duration) == expected


def test_duration_decode_raises_1():
    duration = "PT02H00M42"
    with pytest.raises(ValueError):
        Duration.decode(duration)


def test_duration_decode_raises_2():
    duration = "x-PT02H00M00S"
    with pytest.raises(ValueError):
        Duration.decode(duration)


def test_bool_encode():
    assert Boolean.encode(True) == "true"
    assert Boolean.encode(False) == "false"
    assert Boolean.encode("true") == "true"
    assert Boolean.encode("false") == "false"


def test_bool_bad_encode_on():
    with pytest.raises(TypeError):
        Boolean.encode("on")


def test_bool_encode_bytes():
    assert Boolean.encode(b"True") == "true"


def test_bool_bad_encode_one():
    with pytest.raises(TypeError):
        Boolean.encode(1)


def test_bool_decode():
    assert Boolean.decode("true") is True
    assert Boolean.decode("false") is False


def test_bool_bad_decode_true():
    with pytest.raises(ValueError):
        Boolean.decode("True")


def test_bool_bad_encode_pne():
    with pytest.raises(ValueError):
        Boolean.decode("1")


def test_str_unit_compatibility():
    unit = Unit("1.847mm")
    assert unit.value == Decimal("1.847")
    assert unit.unit == "mm"



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>

import pytest

from odfdo.style import hex2rgb


def test_color_low():
    color = "#012345"
    expected = (1, 35, 69)
    assert hex2rgb(color) == expected


def test_color_high():
    color = "#ABCDEF"
    expected = (171, 205, 239)
    assert hex2rgb(color) == expected


def test_color_lowercase():
    color = "#abcdef"
    expected = (171, 205, 239)
    assert hex2rgb(color) == expected


def test_color_bad_size():
    color = "#fff"
    with pytest.raises(ValueError):
        hex2rgb(color)


def test_color_bad_format():
    color = "978EAE"
    with pytest.raises(ValueError):
        hex2rgb(color)


def test_color_bad_hex():
    color = "#978EAZ"
    with pytest.raises(ValueError):
        hex2rgb(color)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
"""Annotation class for "office:annotation" tag."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Union, cast

from .element import Element, PropDef, register_element_class
from .elements_between import elements_between
from .mixin_dc_creator import DcCreatorMixin
from .mixin_dc_date import DcDateMixin
from .mixin_link import LinkMixin
from .mixin_list import ListMixin
from .mixin_md import MDTail

if TYPE_CHECKING:
    from .body import Body


class AnnotationMixin(Element):
    """Mixin class for classes containing Annotations.

    Used by the following classes:
        - "table:covered-table-cell"
        - "table:table-cell"
        - "text:a"
        - "text:h"
        - "text:meta"
        - "text:meta-field"
        - "text:p"
        - "text:ruby-base"
        - "text:span"

    and "office:text", "office:spreadsheet" for compatibility with previous
    versions.
    """

    def get_annotations(
        self,
        creator: str | None = None,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        content: str | None = None,
    ) -> list[Annotation]:
        """Return all the annotations that match the criteria.

        Args:
            creator: The creator of the annotation.
            start_date: The start date for filtering.
            end_date: The end date for filtering.
            content: A regex to match in the content.

        Returns:
            list[Annotation]: A list of matching Annotation elements.
        """
        annotations: list[Annotation] = []
        for annotation in cast(
            list[Annotation],
            self._filtered_elements("descendant::office:annotation", content=content),
        ):
            if creator is not None and creator != annotation.dc_creator:
                continue
            date = annotation.date
            # date never None: recreated if missing
            if date is None:  # pragma: no cover
                continue
            if start_date is not None and date < start_date:
                continue
            if end_date is not None and date >= end_date:
                continue
            annotations.append(annotation)
        return annotations

    def get_annotation(
        self,
        position: int = 0,
        creator: str | None = None,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        content: str | None = None,
        name: str | None = None,
    ) -> Annotation | None:
        """Return the annotation that matches the criteria.

        Args:
            position: The position of the annotation to return.
            creator: The creator of the annotation.
            start_date: The start date for filtering.
            end_date: The end date for filtering.
            content: A regex to match in the content.
            name: The name of the annotation.

        Returns:
            Annotation or None: The matching Annotation element, or None if not found.
        """
        if name is not None:
            return cast(
                Union[None, Annotation],
                self._filtered_element(
                    "descendant::office:annotation", 0, office_name=name
                ),
            )
        annotations: list[Annotation] = cast(
            list[Annotation],
            self.get_annotations(
                creator=creator,
                start_date=start_date,
                end_date=end_date,
                content=content,
            ),
        )
        if not annotations:
            return None
        try:
            return annotations[position]
        except IndexError:
            return None

    def get_annotation_ends(self) -> list[AnnotationEnd]:
        """Return all the annotation ends.

        Returns:
            list[AnnotationEnd]: A list of AnnotationEnd elements.
        """
        return cast(
            list[AnnotationEnd],
            self._filtered_elements(
                "descendant::office:annotation-end",
            ),
        )

    def get_annotation_end(
        self,
        position: int = 0,
        name: str | None = None,
    ) -> AnnotationEnd | None:
        """Return the annotation end that matches the criteria.

        Args:
            position: The position of the annotation end to return.
            name: The name of the annotation end.

        Returns:
            AnnotationEnd or None: The matching AnnotationEnd element, or
                None if not found.
        """
        return cast(
            Union[None, AnnotationEnd],
            self._filtered_element(
                "descendant::office:annotation-end", position, office_name=name
            ),
        )


def get_unique_office_name(element: Element | None = None) -> str:
    """Provide an autogenerated unique "office:name" for the document.

    Args:
        element: The element to which the annotation is related.

    Returns:
        str: A unique name.
    """
    if element is not None:
        body = element.document_body
    else:
        body = None
    if body:
        used = set(body.get_office_names())
    else:
        used = set()
    # unplugged current paragraph:
    if element is not None:
        used.update(element.get_office_names())
    indice = 1
    while True:
        name = f"__Fieldmark__lpod_{indice}"
        if name in used:
            indice += 1
            continue
        break
    return name


class Annotation(MDTail, ListMixin, LinkMixin, Element, DcCreatorMixin, DcDateMixin):
    """An annotation (private note), "office:annotation".

    This element contains the content of a comment or annotation, along with
    metadata such as the creator and date.

    Attributes:
        name (str): The unique name of the annotation.
        note_id (str): The ID of the note.
        creator (str): The creator of the annotation.
        date (datetime): The date of the annotation.
        note_body (str or Element): The body content of the annotation.
    """

    _tag = "office:annotation"
    _properties = (
        PropDef("name", "office:name"),
        PropDef("note_id", "text:id"),
    )

    def __init__(
        self,
        text_or_element: Element | str | None = None,
        creator: str | None = None,
        date: datetime | None = None,
        name: str | None = None,
        parent: Element | None = None,
        **kwargs: Any,
    ) -> None:
        """Create an Annotation element.

        An annotation is a private note, represented by "office:annotation".
        This element is credited to a creator and contains text content,
        optionally dated (current date by default).

        If 'name' is not provided and a 'parent' is given, the name will be
        autogenerated to be unique within the document.

        Args:
            text_or_element: The content of the annotation, which can be a
                string or another ODF element.
            creator: The name of the person who created the annotation.
            date: The date and time when the annotation was created. Defaults
                to the current time if not provided.
            name: A unique name for the annotation. If not provided, a unique
                name is generated.
            parent: The parent element to which this annotation will be
                associated for name generation.
        """
        # fixme : use offset
        # TODO allow paragraph and text styles
        super().__init__(**kwargs)

        if self._do_init:
            self.note_body = text_or_element
            if creator:
                self.creator = creator
            if date is None:
                date = datetime.now()
            self.date = date
            if not name:
                name = get_unique_office_name(parent)
            self.name = name

    @property
    def dc_creator(self) -> str | None:
        """Alias for self.creator property."""
        return self.creator

    @dc_creator.setter
    def dc_creator(self, creator: str) -> None:
        self.creator = creator

    @property
    def dc_date(self) -> datetime | None:
        """Alias for self.date property."""
        return self.date

    @dc_date.setter
    def dc_date(self, dtdate: datetime) -> None:
        self.date = dtdate

    @property
    def note_body(self) -> str:
        return self.text_content

    @note_body.setter
    def note_body(self, text_or_element: Element | str | None) -> None:
        if text_or_element is None:
            self.text_content = ""
        elif isinstance(text_or_element, str):
            self.text_content = text_or_element
        elif isinstance(text_or_element, Element):
            self.clear()
            self.append(text_or_element)
        else:
            raise TypeError(f'Unexpected type for body: "{type(text_or_element)}"')

    @property
    def start(self) -> Annotation:
        """Return self."""
        return self

    @property
    def end(self) -> AnnotationEnd | None:
        """Return the corresponding annotation-end tag or None."""
        name = self.name
        parent = self.parent
        if parent is None:  # pragma: nocover
            raise ValueError("Can't find end tag: no parent available")
        body: Body | Element = self.document_body or parent
        if hasattr(body, "get_annotation_end"):
            return cast(Union[None, AnnotationEnd], body.get_annotation_end(name=name))
        return None

    def get_annotated(
        self,
        as_text: bool = False,
        no_header: bool = True,
        clean: bool = True,
    ) -> Element | list | str | None:
        """Returns the annotated content from an annotation.

        If no content exists (e.g., single position annotation or if the
        annotation-end tag is not found), it returns an empty list or an
        empty string if 'as_text' is True.

        Args:
            as_text: If True, returns the text content as a string. Defaults
                to False.
            no_header: If True, converts 'text:h' elements to 'text:p'.
                Defaults to True.
            clean: If True, suppresses unwanted tags such as deletion marks.
                Defaults to True.

        Returns:
            Element | list | str | None: The content of the annotation, which
            can be a single element, a list of elements, a string, or None if
            the end tag is missing.
        """
        end = self.end
        if end is None:
            if as_text:
                return ""
            return None
        body: Body | Element = self.document_body or self.root
        return elements_between(
            body, self, end, as_text=as_text, no_header=no_header, clean=clean
        )

    def delete(self, child: Element | None = None, keep_tail: bool = True) -> None:
        """Delete the element from the XML tree.

        If a child element is provided, it is deleted. If no child is given,
        the annotation itself and its corresponding "annotation-end" tag
        (if it exists) are deleted.

        Args:
            child: The child element to delete. If None,
                the annotation itself is deleted. Defaults to None.
            keep_tail: This argument is not used in this context but is
                kept for compatibility. Defaults to True.
        """
        if child is not None:  # act like normal delete
            super().delete(child)
            return
        end = self.end
        if end:
            end.delete()
        # act like normal delete
        super().delete()

    def check_validity(self) -> None:
        """Checks the validity of the Annotation."""
        if not self.note_body:
            raise ValueError("Annotation must have a body")
        if not self.dc_creator:
            raise ValueError("Annotation must have a creator")
        if not self.dc_date:
            self.dc_date = datetime.now()

    def __str__(self) -> str:
        return f"{self.note_body}\n{self.dc_creator} {self.dc_date}"


Annotation._define_attribut_property()


class AnnotationEnd(MDTail, Element):
    """End of annotation marker, "office:annotation-end".

    The "office:annotation-end" element is used to define the end of a text
    range that is being annotated, especially when the content spans across
    element boundaries. It must be preceded by an "office:annotation" element
    with the same 'office:name' attribute.

    If an "office:annotation-end" element does not have a matching preceding
    "office:annotation" element, it is ignored.

    Attributes:
        name (str): The name of the annotation this element is closing. This
                    name must match the 'office:name' of a preceding
                    "office:annotation" element.
    """

    _tag = "office:annotation-end"
    _properties = (PropDef("name", "office:name"),)

    def __init__(
        self,
        annotation: Annotation | None = None,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create an AnnotationEnd element.

        This element marks the end of an annotation. It must be associated
        with an existing "office:annotation" element, either by passing the
        annotation object directly or by providing a matching name.

        Args:
            annotation: The "office:annotation" element
                that this "annotation-end" is closing. If provided, the 'name'
                attribute will be taken from this annotation.
            name: The name of the annotation to close. This is
                required if 'annotation' is not provided.
        """
        # fixme : use offset
        # TODO allow paragraph and text styles
        super().__init__(**kwargs)
        if self._do_init:
            if annotation:
                name = annotation.name
            if not name:
                raise ValueError("Annotation-end must have a name")
            self.name = name

    @property
    def start(self) -> Annotation | None:
        """Return the corresponding annotation starting tag or None."""
        name = self.name
        parent = self.parent
        if parent is None:
            raise ValueError(
                "Can't find start tag: no parent available"
            )  # pragma: nocover
        body: Body | Element = self.document_body or parent
        if hasattr(body, "get_annotation"):  # pragma: nocover
            return cast(Union[None, Annotation], body.get_annotation(name=name))
        return None  # pragma: nocover

    @property
    def end(self) -> AnnotationEnd:
        """Return self."""
        return self


AnnotationEnd._define_attribut_property()

register_element_class(Annotation)
register_element_class(AnnotationEnd)



import ast
import os
import shutil
import subprocess as sp
from pathlib import Path

DOC_SRC = Path(__file__).parent
ROOT = DOC_SRC.parent
SRC = DOC_SRC / "src"
RECIPES_MD = SRC / "recipes.md"
SITE = DOC_SRC / "site"
DEST = DOC_SRC / ".." / "docs"
RECIPES = ROOT / "recipes"


def clean_dirs() -> None:
    shutil.rmtree(DEST, ignore_errors=True)
    shutil.rmtree(SITE, ignore_errors=True)
    shutil.copy(ROOT / "odfdo.png", SRC)
    shutil.copy(ROOT / "README.md", SRC)
    shutil.copy(ROOT / "CHANGES.md", SRC)


def write_recipes_header() -> None:
    header = (
        "# Recipes\n\n"
        "Recipes source code is in the `/recipes` directory "
        "of `odfdo` sources.\n"
        "Most recipes are autonomous scripts doing actual modifications "
        "of ODF sample files, you can check the results in the "
        "`recipes/recipes_output` directory.\n\n"
    )
    RECIPES_MD.write_text(header)


def get_seq_docstring(path: Path) -> tuple[int, str]:
    sequence = 0
    docstring = ""
    tree = ast.parse(path.read_text())
    for node in tree.body:
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == "_DOC_SEQUENCE":
                    sequence = int(ast.literal_eval(node.value))
                    break
    if (
        isinstance(tree.body[0], ast.Expr)
        and isinstance(tree.body[0].value, ast.Constant)
        and isinstance(tree.body[0].value.value, str)
    ):
        docstring = tree.body[0].value.value
    return sequence, docstring


def recipes_sequence() -> dict[int, tuple[Path, str]]:
    recipes: dict[int, tuple[Path, str]] = {}
    for path in RECIPES.glob("*.py"):
        sequence, docstring = get_seq_docstring(path)
        if not sequence:
            sequence = 5000 + len(recipes)
        while sequence in recipes:
            sequence += 1
        recipes[sequence] = (path, docstring)
    return recipes


def add_recipe_blocks(recipes: dict[int, tuple[Path, str]]) -> None:
    with RECIPES_MD.open("a") as file:
        for key in sorted(recipes.keys()):
            path, docstring = recipes[key]
            print("recipe:", path.stem)
            file.write(f"## {path.stem.replace('_', ' ').capitalize()}\n\n")
            file.write(f"{docstring}\n\n")
            file.write(f'??? code "recipes/{path.name}"\n')
            file.write("    ```python\n")
            file.write(f'    {{% include "../../recipes/{path.name}" %}}\n')
            file.write("    ```\n\n")


def make_doc() -> None:
    env = dict(os.environ)
    env["NO_MKDOCS_2_WARNING"] = "1"
    sp.run(["mkdocs", "build"], cwd=DOC_SRC, env=env)  # noqa: S607
    shutil.move(SITE, DEST)


def main() -> None:
    clean_dirs()
    recipes = recipes_sequence()
    write_recipes_header()
    add_recipe_blocks(recipes)
    make_doc()


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
"""Style-related constants and mappings for ODF documents.

This module defines constants and dictionaries that map style family names
to their corresponding ODF tag names, as specified in the ODF 1.2 standard.
It is used internally to manage and register different types of styles.
"""

from __future__ import annotations

# style:family as defined by ODF 1.2, e.g. xxx possibily for:
# 'style:style style:family="xxx"'
FAMILY_ODF_STD = {
    "chart",
    "drawing-page",
    "graphic",
    "paragraph",
    "presentation",
    "ruby",
    "section",
    "table",
    "table-cell",
    "table-column",
    "table-row",
    "text",
}

_BASE_FAMILY_MAP = {k: "style:style" for k in FAMILY_ODF_STD}  # noqa:C420

_FALSE_FAMILY_MAP = {
    "background-image": "style:background-image",
    "date": "number:date-style",
    "font-face": "style:font-face",
    "list": "text:list-style",
    "marker": "draw:marker",
    "number": "number:number-style",
    "outline": "text:outline-style",
    "page-layout": "style:page-layout",
    "percentage": "number:percentage-style",
    "presentation-page-layout": "style:presentation-page-layout",
    "time": "number:time-style",
    "boolean": "number:boolean-style",
    "currency": "number:currency-style",
    "master-page": "style:master-page",
}
# "tab-stop": "style:tab-stop", Defined in toc.py
# "master-page": "style:master-page",


OTHER_STYLES = {
    "style:default-style",
    "style:footer-style",
    "style:header-style",
    "text:list-level-style-bullet",
    "text:list-level-style-image",
    "text:list-level-style-number",
}

SUBCLASS_STYLES = {"background-image"}

FAMILY_MAPPING = {**_BASE_FAMILY_MAP, **_FALSE_FAMILY_MAP}

FALSE_FAMILY_MAP_REVERSE = {v: k for k, v in _FALSE_FAMILY_MAP.items()}
SUBCLASSED_STYLES = {
    "style:background-image",
    "style:master-page",
    "style:page-layout",
    "draw:marker",
}
STYLES_TO_REGISTER = (set(FAMILY_MAPPING.values()) | OTHER_STYLES) - SUBCLASSED_STYLES

# This mapping is not exhaustive, it only contains cases where
# replacing '_' with '-' and adding the "fo:" prefix is not enough
_BASE_PROPERTY_MAPPING = {  # text
    "family_generic": "style:font-family-generic",
    "font": "style:font-name",
    "outline": "style:text-outline",
    "pitch": "style:font-pitch",
    "size": "fo:font-size",
    "style": "fo:font-style",
    "underline": "style:text-underline-style",
    "weight": "fo:font-weight",
    # compliance with office suites
    "font_family": "fo:font-family",
    "font_style_name": "style:font-style-name",
    # paragraph
    "align-last": "fo:text-align-last",
    "align": "fo:text-align",
    "indent": "fo:text-indent",
    "together": "fo:keep-together",
    # style:paragraph-properties : built
    # frame position
    "horizontal_pos": "style:horizontal-pos",
    "horizontal_rel": "style:horizontal-rel",
    "vertical_pos": "style:vertical-pos",
    "vertical_rel": "style:vertical-rel",
    # TODO 'page-break-before': 'fo:page-break-before',
    # TODO 'page-break-after': 'fo:page-break-after',
    # Graphic
    "fill_color": "draw:fill-color",
    "fill_image_height": "draw:fill-image-height",
    "fill_image_width": "draw:fill-image-width",
    "guide_distance": "draw:guide-distance",
    "guide_overhang": "draw:guide-overhang",
    "line_distance": "draw:line-distance",
    "stroke": "draw:stroke",
    "textarea_vertical_align": "draw:textarea-vertical-align",
}

STYLE_ATTRIBUTES = {
    # 17.6<style:paragraph-properties>
    "paragraph": {
        "fo:background-color",
        "fo:border",
        "fo:border-bottom",
        "fo:border-left",
        "fo:border-right",
        "fo:border-top",
        "fo:break-after",
        "fo:break-before",
        "fo:hyphenation-keep",
        "fo:hyphenation-ladder-count",
        "fo:keep-together",
        "fo:keep-with-next",
        "fo:line-height",
        "fo:margin",
        "fo:margin-bottom",
        "fo:margin-left",
        "fo:margin-right",
        "fo:margin-top",
        "fo:orphans",
        "fo:padding",
        "fo:padding-bottom",
        "fo:padding-left",
        "fo:padding-right",
        "fo:padding-top",
        "fo:text-align",
        "fo:text-align-last",
        "fo:text-indent",
        "fo:widows",
        "style:auto-text-indent",
        "style:background-transparency",
        "style:border-line-width",
        "style:border-line-width-bottom",
        "style:border-line-width-left",
        "style:border-line-width-right",
        "style:border-line-width-top",
        "style:font-independent-line-spacing",
        "style:join-border",
        "style:justify-single-word",
        "style:line-break",
        "style:line-height-at-least",
        "style:line-spacing",
        "style:page-number",
        "style:punctuation-wrap",
        "style:register-true",
        "style:shadow",
        "style:snap-to-layout-grid",
        "style:tab-stop-distance",
        "style:text-autospace",
        "style:vertical-align",
        "style:writing-mode",
        "style:writing-mode-automatic",
        "text:line-number",
        "text:number-lines",
    },
    # 16.27.28<style:text-properties>
    "text": {
        "fo:background-color",
        "fo:color",
        "fo:country",
        "fo:font-family",
        "fo:font-size",
        "fo:font-style",
        "fo:font-variant",
        "fo:font-weight",
        "fo:hyphenate",
        "fo:hyphenation-push-char-count",
        "fo:hyphenation-remain-char-count",
        "fo:language",
        "fo:letter-spacing",
        "fo:script",
        "fo:text-shadow",
        "fo:text-transform",
        "style:country-asian",
        "style:country-complex",
        "style:font-charset",
        "style:font-charset-asian",
        "style:font-charset-complex",
        "style:font-family-asian",
        "style:font-family-complex",
        "style:font-family-generic",
        "style:font-family-generic-asian",
        "style:font-family-generic-complex",
        "style:font-name",
        "style:font-name-asian",
        "style:font-name-complex",
        "style:font-pitch",
        "style:font-pitch-asian",
        "style:font-pitch-complex",
        "style:font-relief",
        "style:font-size-asian",
        "style:font-size-complex",
        "style:font-size-rel",
        "style:font-size-rel-asian",
        "style:font-size-rel-complex",
        "style:font-style-asian",
        "style:font-style-complex",
        "style:font-style-name",
        "style:font-style-name-asian",
        "style:font-style-name-complex",
        "style:font-weight-asian",
        "style:font-weight-complex",
        "style:language-asian",
        "style:language-complex",
        "style:letter-kerning",
        "style:rfc-language-tag",
        "style:rfc-language-tag-asian",
        "style:rfc-language-tag-complex",
        "style:script-asian",
        "style:script-complex",
        "style:script-type",
        "style:text-blinking",
        "style:text-combine",
        "style:text-combine-end-char",
        "style:text-combine-start-char",
        "style:text-emphasize",
        "style:text-line-through-color",
        "style:text-line-through-mode",
        "style:text-line-through-style",
        "style:text-line-through-text",
        "style:text-line-through-text-style",
        "style:text-line-through-type",
        "style:text-line-through-width",
        "style:text-outline",
        "style:text-overline-color",
        "style:text-overline-mode",
        "style:text-overline-style",
        "style:text-overline-type",
        "style:text-overline-width",
        "style:text-position",
        "style:text-rotation-angle",
        "style:text-rotation-scale",
        "style:text-scale",
        "style:text-underline-color",
        "style:text-underline-mode",
        "style:text-underline-style",
        "style:text-underline-type",
        "style:text-underline-width",
        "style:use-window-font-color",
        "text:condition",
        "text:display",
    },
    "page-layout": {
        "fo:background-color",
        "fo:border",
        "fo:border-bottom",
        "fo:border-left",
        "fo:border-right",
        "fo:border-top",
        "fo:margin",
        "fo:margin-bottom",
        "fo:margin-left",
        "fo:margin-right",
        "fo:margin-top",
        "fo:padding",
        "fo:padding-bottom",
        "fo:padding-left",
        "fo:padding-right",
        "fo:padding-top",
        "fo:page-height",
        "fo:page-width",
        "style:border-line-width",
        "style:border-line-width-bottom",
        "style:border-line-width-left",
        "style:border-line-width-right",
        "style:border-line-width-top",
        "style:first-page-number",
        "style:footnote-max-height",
        "style:footnote-sep",
        "style:layout-grid-base-height",
        "style:layout-grid-base-width",
        "style:layout-grid-color",
        "style:layout-grid-display",
        "style:layout-grid-lines",
        "style:layout-grid-mode",
        "style:layout-grid-print",
        "style:layout-grid-ruby-below",
        "style:layout-grid-ruby-height",
        "style:layout-grid-snap-to",
        "style:layout-grid-standard-mode",
        "style:num-format",
        "style:num-letter-sync",
        "style:num-prefix",
        "style:num-suffix",
        "style:paper-tray-name",
        "style:print",
        "style:print-orientation",
        "style:print-page-order",
        "style:register-truth-ref-style-name",
        "style:scale-to",
        "style:scale-to-pages",
        "style:shadow",
        "style:table-centering",
        "style:writing-mode",
    },
    # 17.10<style:ruby-properties>
    "ruby": {"style:ruby-align", "style:ruby-position"},
    # 17.11<style:section-properties>
    "section": {
        "fo:background-color",
        "fo:margin-left",
        "fo:margin-right",
        "style:editable",
        "style:protect",
        "style:writing-mode",
        "text:dont-balance-text-columns",
    },
    # 17.15<style:table-properties>
    "table": {
        "fo:background-color",
        "fo:break-after",
        "fo:break-before",
        "fo:keep-with-next",
        "fo:margin",
        "fo:margin-bottom",
        "fo:margin-left",
        "fo:margin-right",
        "fo:margin-top",
        "style:may-break-between-rows",
        "style:page-number",
        "style:rel-width",
        "style:shadow",
        "style:width",
        "style:writing-mode",
        "table:align",
        "table:border-model",
        "table:display",
    },
    # 17.16<style:table-column-properties>
    "table-column": {
        "fo:break-after",
        "fo:break-before",
        "style:column-width",
        "style:rel-column-width",
        "style:use-optimal-column-width",
    },
    # 17.17<style:table-row-properties>
    "table-row": {
        "fo:background-color",
        "fo:break-after",
        "fo:break-before",
        "fo:keep-together",
        "style:min-row-height",
        "style:row-height",
        "style:use-optimal-row-height",
    },
    # 17.18<style:table-cell-properties>
    "table-cell": {
        "fo:background-color",
        "fo:border",
        "fo:border-bottom",
        "fo:border-left",
        "fo:border-right",
        "fo:border-top",
        "fo:padding",
        "fo:padding-bottom",
        "fo:padding-left",
        "fo:padding-right",
        "fo:padding-top",
        "fo:wrap-option",
        "style:border-line-width",
        "style:border-line-width-bottom",
        "style:border-line-width-left",
        "style:border-line-width-right",
        "style:border-line-width-top",
        "style:cell-protect",
        "style:decimal-places",
        "style:diagonal-bl-tr",
        "style:diagonal-bl-tr-widths",
        "style:diagonal-tl-br",
        "style:diagonal-tl-br-widths",
        "style:direction",
        "style:glyph-orientation-vertical",
        "style:print-content",
        "style:repeat-content",
        "style:rotation-align",
        "style:rotation-angle",
        "style:shadow",
        "style:shrink-to-fit",
        "style:text-align-source",
        "style:vertical-align",
        "style:writing-mode",
    },
    # 17.21<style:graphic-properties>
    "graphic": {
        "dr3d:ambient-color",
        "dr3d:backface-culling",
        "dr3d:back-scale",
        "dr3d:close-back",
        "dr3d:close-front",
        "dr3d:depth",
        "dr3d:diffuse-color",
        "dr3d:edge-rounding",
        "dr3d:edge-rounding-mode",
        "dr3d:emissive-color",
        "dr3d:end-angle",
        "dr3d:horizontal-segments",
        "dr3d:lighting-mode",
        "dr3d:normals-direction",
        "dr3d:normals-kind",
        "dr3d:shadow",
        "dr3d:shininess",
        "dr3d:specular-color",
        "dr3d:texture-filter",
        "dr3d:texture-generation-mode-x",
        "dr3d:texture-generation-mode-y",
        "dr3d:texture-kind",
        "dr3d:texture-mode",
        "dr3d:vertical-segments",
        "draw:auto-grow-height",
        "draw:auto-grow-width",
        "draw:blue",
        "draw:caption-angle",
        "draw:caption-angle-type",
        "draw:caption-escape",
        "draw:caption-escape-direction",
        "draw:caption-fit-line-length",
        "draw:caption-gap",
        "draw:caption-line-length",
        "draw:caption-type",
        "draw:color-inversion",
        "draw:color-mode",
        "draw:contrast",
        "draw:decimal-places",
        "draw:draw-aspect",
        "draw:end-guide",
        "draw:end-line-spacing-horizontal",
        "draw:end-line-spacing-vertical",
        "draw:fill",
        "draw:fill-color",
        "draw:fill-gradient-name",
        "draw:fill-hatch-name",
        "draw:fill-hatch-solid",
        "draw:fill-image-height",
        "draw:fill-image-name",
        "draw:fill-image-ref-point",
        "draw:fill-image-ref-point-x",
        "draw:fill-image-ref-point-y",
        "draw:fill-image-width",
        "draw:fit-to-contour",
        "draw:fit-to-size",
        "draw:frame-display-border",
        "draw:frame-display-scrollbar",
        "draw:frame-margin-horizontal",
        "draw:frame-margin-vertical",
        "draw:gamma",
        "draw:gradient-step-count",
        "draw:green",
        "draw:guide-distance",
        "draw:guide-overhang",
        "draw:image-opacity",
        "draw:line-distance",
        "draw:luminance",
        "draw:marker-end",
        "draw:marker-end-center",
        "draw:marker-end-width",
        "draw:marker-start",
        "draw:marker-start-center",
        "draw:marker-start-width",
        "draw:measure-align",
        "draw:measure-vertical-align",
        "draw:ole-draw-aspect",
        "draw:opacity",
        "draw:opacity-name",
        "draw:parallel",
        "draw:placing",
        "draw:red",
        "draw:secondary-fill-color",
        "draw:shadow",
        "draw:shadow-color",
        "draw:shadow-offset-x",
        "draw:shadow-offset-y",
        "draw:shadow-opacity",
        "draw:show-unit",
        "draw:start-guide",
        "draw:start-line-spacing-horizontal",
        "draw:start-line-spacing-vertical",
        "draw:stroke",
        "draw:stroke-dash",
        "draw:stroke-dash-names",
        "draw:stroke-linejoin",
        "draw:symbol-color",
        "draw:textarea-horizontal-align",
        "draw:textarea-vertical-align",
        "draw:tile-repeat-offset",
        "draw:unit",
        "draw:visible-area-height",
        "draw:visible-area-left",
        "draw:visible-area-top",
        "draw:visible-area-width",
        "draw:wrap-influence-on-position",
        "fo:background-color",
        "fo:border",
        "fo:border-bottom",
        "fo:border-left",
        "fo:border-right",
        "fo:border-top",
        "fo:clip",
        "fo:margin",
        "fo:margin-bottom",
        "fo:margin-left",
        "fo:margin-right",
        "fo:margin-top",
        "fo:max-height",
        "fo:max-width",
        "fo:min-height",
        "fo:min-width",
        "fo:padding",
        "fo:padding-bottom",
        "fo:padding-left",
        "fo:padding-right",
        "fo:padding-top",
        "fo:wrap-option",
        "style:background-transparency",
        "style:border-line-width",
        "style:border-line-width-bottom",
        "style:border-line-width-left",
        "style:border-line-width-right",
        "style:border-line-width-top",
        "style:editable",
        "style:flow-with-text",
        "style:horizontal-pos",
        "style:horizontal-rel",
        "style:mirror",
        "style:number-wrapped-paragraphs",
        "style:overflow-behavior",
        "style:print-content",
        "style:protect",
        "style:rel-height",
        "style:rel-width",
        "style:repeat",
        "style:run-through",
        "style:shadow",
        "style:shrink-to-fit",
        "style:vertical-pos",
        "style:vertical-rel",
        "style:wrap",
        "style:wrap-contour",
        "style:wrap-contour-mode",
        "style:wrap-dynamic-threshold",
        "style:writing-mode",
        "svg:fill-rule",
        "svg:height",
        "svg:stroke-color",
        "svg:stroke-linecap",
        "svg:stroke-opacity",
        "svg:stroke-width",
        "svg:width",
        "svg:x",
        "svg:y",
        "text:anchor-page-number",
        "text:anchor-type",
        "text:animation",
        "text:animation-delay",
        "text:animation-direction",
        "text:animation-repeat",
        "text:animation-start-inside",
        "text:animation-steps",
        "text:animation-stop-inside",
    },
}



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from collections.abc import Iterable
from textwrap import dedent

import pytest

from odfdo.document import Document
from odfdo.paragraph import Paragraph


@pytest.fixture
def document_base(samples) -> Iterable[Document]:
    document = Document(samples("base_text.odt"))
    yield document


@pytest.fixture
def document_example(samples) -> Iterable[Document]:
    document = Document(samples("example.odt"))
    yield document


@pytest.fixture
def document_bookmark(samples) -> Iterable[Document]:
    document = Document(samples("bookmark.odt"))
    yield document


@pytest.fixture
def document_i28(samples) -> Iterable[Document]:
    document = Document(samples("issue_28_pretty.odt"))
    yield document


@pytest.fixture
def document_list(samples) -> Iterable[Document]:
    document = Document(samples("list.odt"))
    yield document


@pytest.fixture
def document_meta(samples) -> Iterable[Document]:
    document = Document(samples("meta.odt"))
    yield document


@pytest.fixture
def document_note(samples) -> Iterable[Document]:
    document = Document(samples("note.odt"))
    yield document


@pytest.fixture
def document_pb(samples) -> Iterable[Document]:
    document = Document(samples("pagebreak.odt"))
    yield document


@pytest.fixture
def document_toc2(samples) -> Iterable[Document]:
    document = Document(samples("toc_done.odt"))
    yield document


@pytest.fixture
def document_toc(samples) -> Iterable[Document]:
    document = Document(samples("toc.odt"))
    yield document


@pytest.fixture
def document_tc(samples) -> Iterable[Document]:
    document = Document(samples("tracked_changes.odt"))
    yield document


@pytest.fixture
def document_uf(samples) -> Iterable[Document]:
    document = Document(samples("user_fields.odt"))
    yield document


@pytest.fixture
def document_var(samples) -> Iterable[Document]:
    document = Document(samples("variable.odt"))
    yield document


@pytest.fixture
def document_case1(samples) -> Iterable[Document]:
    document = Document(samples("base_md_text.odt"))
    yield document


@pytest.fixture
def document_case2(samples) -> Iterable[Document]:
    document = Document(samples("md_style.odt"))
    yield document


@pytest.fixture
def document_case3(samples) -> Iterable[Document]:
    document = Document(samples("md_fixed.odt"))
    yield document


@pytest.fixture
def document_dormeur(samples) -> Iterable[Document]:
    document = Document(samples("dormeur_notes.odt"))
    yield document


@pytest.fixture
def document_lorem(samples) -> Iterable[Document]:
    document = Document(samples("lorem.odt"))
    yield document


@pytest.fixture
def document_img(samples) -> Iterable[Document]:
    document = Document(samples("chair.odt"))
    yield document


@pytest.fixture
def document_tab(samples) -> Iterable[Document]:
    document = Document(samples("table.odt"))
    yield document


def test_md_doc_ods():
    doc = Document("ods")
    with pytest.raises(NotImplementedError):
        _md = doc.to_markdown()


def test_md_doc_drawing():
    doc = Document("drawing")
    with pytest.raises(NotImplementedError):
        _md = doc.to_markdown()


def test_md_doc_empty():
    doc = Document("odt")
    md = doc.to_markdown()
    assert md == ""


def test_md_doc_minimal():
    text = "some text"
    doc = Document("odt")
    doc.body.clear()
    doc.body.append(Paragraph(text))
    md = doc.to_markdown()
    assert md.strip() == text


def test_md_base_text(document_base):
    md = document_base.to_markdown()
    expected = dedent(
        """\
    # odfdo Test Case Document

    This is the first paragraph.

    This is the second paragraph.

    This is a paragraph with a named style.

    ## Level 2 Title

    This is the first paragraph of the second title.

    This is the last paragraph with diacritical signs: éè

    # First Title of the Second Section

    First paragraph of the second section.

    This is the second paragraph with [an external link](https://github.com/jdum/odfdo) inside.
    """
    ).strip()
    assert md.strip() == expected


def test_md_example_text(document_example):
    md = document_example.to_markdown()
    expected = dedent(
        """\
    # odfdo Test Case Document

    This is the first paragraph.

    This is the second paragraph.

    This is a paragraph with a named style.

    ## Level 2 Title

    This is the first paragraph of the second title.

    This is the last paragraph with diacritical signs: éè

    # First Title of the Second Section

    First paragraph of the second section.
    """
    ).strip()
    # First paragraphAuteur inconnu2009-06-22T17:18:42This is an annotation.With diacritical signs: éè
    assert md.strip() == expected


def test_md_bookmark_text(document_bookmark):
    md = document_bookmark.to_markdown()
    expected = "Lettre à Élise (cf. page 1)."
    # First paragraphAuteur inconnu2009-06-22T17:18:42This is an annotation.With diacritical signs: éè
    assert md.strip() == expected


def test_md_i28_text(document_i28):
    md = document_i28.to_markdown()
    expected = (
        "This is an example with => v8.1.4 <= spaces after "
        "reading and writing with odfdo."
    )
    # First paragraphAuteur inconnu2009-06-22T17:18:42This is an annotation.With diacritical signs: éè
    assert md.strip() == expected


def test_md_list_text(document_list):
    md = document_list.to_markdown()
    expected = dedent(
        """\
        Some text

         -  une liste accentuée
            -  un sous\\-élément

         1. une liste numérotée
         2. et de deux !
    """
    ).strip()
    # First paragraphAuteur inconnu2009-06-22T17:18:42This is an annotation.With diacritical signs: éè
    assert md.strip() == expected


def test_md_meta_text(document_meta):
    md = document_meta.to_markdown()
    expected = "Only testing meta.xml here…"
    assert md.strip() == expected


def test_md_note_text(document_note):
    md = document_note.to_markdown()
    expected = dedent(
        """\
    Un paragraphe[1] d'apparence[i] banale.

    1. C'est-à-dire l'élément « text:p ».

    i. Les apparences sont trompeuses !
    """  # noqa: RUF001
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected


def test_md_pagebreak_text(document_pb):
    md = document_pb.to_markdown()
    expected = dedent(
        """\
    first paragraph

    second paragraph
    """
    ).strip()
    assert md.strip() == expected


def test_md_toc_text(document_toc):
    md = document_toc.to_markdown()
    expected = dedent(
        """\
    # Level 1 title 1

    ## Level 2 title 1

    # Level 1 title 2

    ### Level 3 title 1

    ## Level 2 title 2

    # Level 1 title 3

    ## Level 2 title 1

    ### Level 3 title 1

    ### Level 3 title 2

    ## Level 2 title 2

    ### Level 3 title 1

    ### Level 3 title 2

    """
    ).strip()
    assert md.strip() == expected


def test_user_fields_text(document_uf):
    md = document_uf.to_markdown()
    expected = dedent(
        """\
    A document with user\\-field declarations.

    Paris

    77 77 77 77 77

    42
    """
    ).strip()
    assert md.strip() == expected


def test_md_variable_text(document_var):  # FAIL
    md = document_var.to_markdown()
    expected = dedent(
        """\
    Today is 12/08/09 15:17:27.

    This document is named Intitulé.

    This is page 1 out of 1.

    Reference to page 1.

    Declaring variable 123.

    Inserting variable 123.

    Inserting a user field VRAI.
    """
    ).strip()
    assert md.strip() == expected


def test_md_case1_text(document_case1):  # FAIL
    md = document_case1.to_markdown()
    expected = dedent(
        """\
        # odfdo Test Case Document

        This is the first paragraph with a link: [the odfdo project](https://github.com/jdum/odfdo).

        This is the second paragraph.

        This is a paragraph with a named style.

        Some list :

         -  item1
         -  item2
         -  item3
            -  sub item a
            -  sub item (with link: [odfdo b link](https://github.com/jdum/odfdo)) b
            -  sub item c
         -  item4

        last paragraph

        ## Level 2 Title with [odfdo project again](https://github.com/jdum/odfdo) embeded

        This is the first paragraph of the second title.

        This is the last paragraph with diacritical signs: éè

        # First Title of the Second Section

        First paragraph of the second section.
    """
    ).strip()
    print(md.strip())
    assert md.strip() == expected


def test_md_case2_text(document_case2):  # FAIL
    # Here  **bolded space** at *begin*.
    # Some style here and *there*

    md = document_case2.to_markdown()
    expected = dedent(
        """\
        # odfdo Test Case2

        Some style `here` and _there_

        Two spaces at start

        ~~somme modified~~

        =7 Starting spaces This is the second paragraph.

        One tab

        Two tabs

        This is a **bold** **paragraph** with a named style.

        Here  **bolded space** at _begin_.

        Some _special_ ***case*** _to_ check if ~~barred~~

        Some list :

         -  item1
         -  **item2**
         -  `item3`
         -  ***item4***
         -  item5

            1. a
            2. b

               -  aaa
               -  eee
                  -  ~~zzzz~~
                  -  zzzz
               -  iii

            3. c

         -  _item6_
         -  [x] done

            1. u
            2. v
            3. w

         -  [ ] undone
         -

        last paragraph _with italic_

        \\#\\#\\#\\# To see **\\*\\*stars\\*\\*** or \\*\\*any\\*\\*

        and **\\*more\\*** and \\~\\~maybe\\~\\~

        ## Level 2 Title with [odfdo project again](https://github.com/jdum/odfdo) embeded

        This is the first paragraph of the second title.

         1. One
         2. **two**
            1. aa
            2. **bb**

               -  eee
               -  fff

            3. **cc**
         3. three
            1. x
            2. y
               1. p
               2. q
               3. r
            3. z
         4. four

        after
        """
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected


def test_md_case3_text(document_case3):  # FAIL
    md = document_case3.to_markdown()
    expected = dedent(
        """\
        # odfdo Test Case3

        ```
        Some text
        ```

        neutral

        ```
        several lines
        of text
        with a
        fixed style
        ```

        standard

         1. a
         2. `here `
         3. `some fixed too`
         4. hop

        ```
        another one,
        with *thing* and other things
        in it
         tabs	in same place
              spaces
        bold  or italic ?
        ```

        End here

        ```


        ```
        """  # noqa: RUF001
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected


def test_md_dormeur_text(document_dormeur):  # FAIL
    md = document_dormeur.to_markdown()
    expected = dedent(
        """\
        # ~~Le~~ dormeur[1] du _val_

        C'est un trou de verdure où chante une rivière,\\
        Accrochant follement aux herbes des haillons\\
        D'argent ; où le soleil, de la _montagne_[2] **fière**,\\
        Luit : c'est un petit val qui mousse de rayons.\\
        \\
        Un soldat jeune, bouche ouverte, tête nue,\\
        Et la nuque baignant dans le frais cresson bleu,\\
        Dort ; il est étendu dans l'herbe, sous la nue,\\
        Pâle dans son lit vert où la lumière pleut.\\
        \\
        Les pieds dans les glaïeuls, il dort. Souriant[i] comme\\
        Sourirait un enfant malade, il fait un somme :\\
        Nature, berce\\-le chaudement : il a froid.\\
        \\
        Les parfums ne font pas frissonner sa narine ;\\
        Il dort dans le soleil, la main sur sa poitrine,\\
        Tranquille. Il a deux trous rouges au côté droit.

        Arthur Rimbaud

        1. Note dormeur
        2. Note next [remote link](https://test.example.com/)

        i. End note

        """
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected


def test_md_lorem_text(document_lorem):  # FAIL
    md = document_lorem.to_markdown()
    expected = dedent(
        """\
        # Lorem ipsum dolor sit amet

        _Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Sed non risus. Suspendisse lectus tortor, dignissim sit amet, adipiscing nec, ultricies sed, dolor. Cras elementum ultrices diam. Maecenas ligula massa, varius a, semper congue, euismod non, mi. Proin porttitor, orci nec nonummy molestie, enim est eleifend mi, non fermentum diam nisl sit amet erat. Duis semper. Duis arcu massa, scelerisque vitae, consequat in,_ **pretium** _a, enim. Pellentesque congue. Ut in risus volutpat libero pharetra tempor. Cras vestibulum bibendum augue. Praesent egestas leo in pede. Praesent blandit odio eu enim. Pellentesque sed dui ut augue blandit sodales. Vestibulum ante ipsum primis in faucibus orci luctus et ultrices posuere cubilia Curae;_ **Aliquam nibh. Mauris ac mauris sed pede pellentesque** _fermentum._

         -  Maecenas adipiscing ante non diam sodales
         -  Ut velit mauris, egestas sed, gravida nec, ornare ut, mi.

        Aenean utorci vel massa suscipit **pulvinar**.

        **Maecenas adipiscing ante non diam sodales hendrerit. Ut velit mauris, egestas sed, gravida nec, ornare ut, mi. Aenean ut orci vel massa suscipit pulvinar. Nulla sollicitudin. Fusce varius, ligula non tempus aliquam, nunc turpis ullamcorper nibh, in tempus sapien eros vitae ligula. Pellentesque rhoncus nunc et augue. Integer id felis. Curabitur aliquet pellentesque diam. Integer quis metus vitae elit lobortis egestas. Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Morbi vel erat non mauris convallis vehicula. Nulla et sapien. Integer tortor tellus, aliquam faucibus, convallis id, congue eu, quam. Mauris ullamcorper felis vitae erat. Proin feugiat, augue non elementum posuere, metus purus iaculis lectus, et tristique ligula justo vitae magna. Aliquam convallis sollicitudin purus. Praesent aliquam, enim at fermentum mollis, ligula massa adipiscing nisl, ac euismod nibh nisl eu lectus. Fusce vulputate sem at sapien. Vivamus leo. Aliquam euismod libero eu enim. Nulla nec felis sed leo placerat imperdiet. Aenean suscipit nulla in justo. Suspendisse cursus rutrum augue. Nulla tincidunt tincidunt mi. Curabitur iaculis, lorem vel rhoncus faucibus, felis magna fermentum augue, et ultricies lacus lorem varius purus.** _Curabitur_ **eu amet.**
        """
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected


def test_md_tracked_text(document_tc):  # FAIL
    md = document_tc.to_markdown()
    expected = dedent(
        """\
        Bonjour  los bonitos amigos  ça va ?
        """
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected


def test_md_toc2_text(document_toc2):
    md = document_toc2.to_markdown()
    expected = dedent(
        """\
        **Table of Contents**

        [Level 1 title 1	1](#)

        [Level 2 title 1	1](#)

        [Level 1 title 2	1](#)

        [Level 3 title 1	1](#)

        [Level 2 title 2	1](#)

        [Level 1 title 3	1](#)

        [Level 2 title 1	1](#)

        [Level 3 title 1	1](#)

        [Level 3 title 2	1](#)

        [Level 2 title 2	1](#)

        [Level 3 title 1	1](#)

        [Level 3 title 2	1](#)

        #

        # Level 1 title 1

        ## Level 2 title 1

        # Level 1 title 2

        ### Level 3 title 1

        ## Level 2 title 2

        # Level 1 title 3

        ## Level 2 title 1

        ### Level 3 title 1

        ### Level 3 title 2

        ## Level 2 title 2

        ### Level 3 title 1

        ### Level 3 title 2
    """
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected


def test_md_img_text(document_img):  # FAIL
    md = document_img.to_markdown()
    expected = dedent(
        """\
        **Image test**

        A chair :

        ![alternate name, chair](Pictures/1000000000000094000001004C83F003.jpg)
        _Figure_ 1  _chair_
        """
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected


def test_md_table_text(document_tab):  # FAIL
    md = document_tab.to_markdown()
    expected = dedent(
        """\
        # First table

        | a                 | b                               | c   | d               |
        |-------------------|---------------------------------|-----|-----------------|
        | Some bar \\| there | Log or _short_ or **very** long |     | \\*\\*no\\*\\* bold |
        | 1                 | 2                               | 3   | 4               |
        | `fixed`           | 20                              | 30  | 40              |
        | 100               | **200**                         | 300 | 400             |


        # Second table

        | AAAAAAAAAAAAAAAAAA | BBBB                   | **CCC**   |                        | EE    |
        |--------------------|------------------------|-----------|------------------------|-------|
        | 1.234              | a                      | _bb_      | 2024\\-12\\-25           | \\-2   |
        | **Some title**     | Some line break inside | anchor[1] | A list of 3 paras here | `123` |


        1. Note in a cell
        """
    ).strip()
    print(repr(md.strip()))
    assert md.strip() == expected



#!/usr/bin/env python
"""Print the metadata informations of a ODF file.

Metadata are accessible through the meta part: meta = document.get_part("meta.xml")
or the shortcut: document.meta.

You then get access to various getters and setters. The getters return
Python types and the respective setters take the same Python type as
a parameter.
"""

import os
import sys
from pathlib import Path

from odfdo import Document

_DOC_SEQUENCE = 490
DATA = Path(__file__).parent / "data"
SOURCE = "collection2.odt"


def read_source_document() -> Document:
    """Return the source Document."""
    try:
        source = sys.argv[1]
    except IndexError:
        source = DATA / SOURCE
    return Document(source)


def print_meta(document: Document) -> None:
    """Print the medadata of the document.

    Output of the get_xxx methods for metadata.
    Notice that odfdo doesn't increment editing cycles nor statistics
    when saving the document.
    For the metadata using dates or durations, odfdo provides datatypes that
    decode from and serialize back to strings.
    Strings are always decoded as utf-8, numeric values are decoded
    as Decimal."""

    meta = document.meta

    print(f"Meta data of {document.path}")
    print("Title                :", meta.title)
    print("creator              :", meta.creator)
    print("creation date        :", meta.creation_date)
    print("modification date    :", meta.date)
    print("initial creator      :", meta.initial_creator)
    print("subject              :", meta.subject)
    print("description          :", meta.description)
    print("editing cycles       :", meta.editing_cycles)
    print("editing duration     :", meta.editing_duration)
    print("generator            :", meta.generator)
    print("language             :", meta.language)
    print("keywords             :", meta.keyword)
    print("statistics    ")
    if meta.statistic is not None:
        for key, value in meta.statistic.items():
            print(f"   {key[5:]:<18}: {value}")
    user_defined = meta.user_defined_metadata
    if user_defined:
        print("user defined metadata")
        for key, value in user_defined.items():
            print(f"   {key[5:]:<18}: {value}")

    # A quick way to have all of those informations:
    print("-" * 70)
    print(document.get_formated_meta())


def main() -> None:
    document = read_source_document()
    print_meta(document)
    test_unit(document)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    meta = document.meta
    assert meta.keyword.strip() == "Douglas Adams"
    assert meta.statistic["meta:page-count"] == 20


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_controls import FormPassword


@pytest.fixture
def form_password() -> Iterable[FormPassword]:
    yield FormPassword(
        name="some name",
        control_implementation="some implem",
        value="initial",
        convert_empty_to_null=False,
        disabled=False,
        printable=True,
        tab_index=4,
        tab_stop=False,
        xml_id="control1",
        max_length=64,
    )


def test_form_password_class():
    form = FormPassword()
    assert isinstance(form, FormPassword)


def test_form_password_minimal_tag():
    form = Element.from_tag("<form:password/>")
    assert isinstance(form, FormPassword)


def test_form_password_serialize():
    form = FormPassword()
    assert form.serialize() == '<form:password form:echo-char="*"/>'


def test_form_password_repr():
    form = FormPassword()
    assert repr(form) == "<FormPassword name=None xml_id=None>"


def test_form_password_name(form_password):
    assert form_password.name == "some name"


def test_form_password_control_implementation(form_password):
    assert form_password.control_implementation == "some implem"


def test_form_password_disabled(form_password):
    assert form_password.disabled is False


def test_form_password_printable(form_password):
    assert form_password.printable is True


def test_form_password_tab_index(form_password):
    assert form_password.tab_index == 4


def test_form_password_tab_stop(form_password):
    assert form_password.tab_stop is False


def test_form_password_xml_id(form_password):
    assert form_password.xml_id == "control1"


def test_form_password_form_id(form_password):
    assert form_password.form_id == "control1"


def test_form_password_max_length(form_password):
    assert form_password.max_length == 64


def test_form_password_xforms_bind():
    form = FormPassword(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        xforms_bind="bind id",
    )
    assert form.xforms_bind == "bind id"


def test_form_password_form_id_2():
    form = FormPassword(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
    )
    assert form.form_id == "control1"


def test_form_password_form_value(form_password):
    assert form_password.value == "initial"


def test_form_password_form_as_dict(form_password):
    expected = {
        "tag": "form:password",
        "name": "some name",
        "xml_id": "control1",
        "value": "initial",
        "current_value": None,
        "str": "*",
    }
    assert form_password.as_dict() == expected


def test_form_password_form_as_dict2(form_password):
    form_password.current_value = None
    expected = {
        "tag": "form:password",
        "name": "some name",
        "xml_id": "control1",
        "value": "initial",
        "current_value": None,
        "str": "*",
    }
    assert form_password.as_dict() == expected


def test_form_password_convert_empty_to_null(form_password):
    assert form_password.convert_empty_to_null is False


def test_form_password_linked_cell(form_password):
    assert form_password.linked_cell is None


def test_form_password_echo_char(form_password):
    assert form_password.echo_char == "*"


def test_form_password_echo_char_2(form_password):
    form_password.echo_char = "X"
    assert form_password.echo_char == "X"


def test_form_password_echo_char_3():
    form = FormPassword(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
        echo_char="X",
    )
    assert form.echo_char == "X"


def test_form_password_echo_char_4():
    form = FormPassword(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
        echo_char="X",
    )
    assert form.echo_char == "X"


def test_form_password_echo_char_5():
    form = FormPassword(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
        echo_char=None,
    )
    assert form.echo_char == "*"


def test_form_password_linked_cell_2():
    form = FormPassword(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
        linked_cell="a1",
    )
    assert form.linked_cell == "a1"
